import logging
from typing import Callable, Iterable

from dataclasses import dataclass
from requests import Session


@dataclass
class DicomReceivedMatcher:
    matches: Callable[[str, Session], bool]
    modality_selector: Callable[[str, Session], str]


def forward_dicom(matchers: Iterable[DicomReceivedMatcher]):
    def forward_series(event, session):
        resource_id = event.resource_id
        for matcher in matchers:
            if matcher.matches(resource_id, session):
                modality = matcher.modality_selector(resource_id, session)
                resp = session.post(f'/modalities/{modality}/store', json=[resource_id])
                resp.raise_for_status()
                logging.info(f"DICOM export to modality '{modality}' started for resource '{resource_id}'")

            else:
                logging.info(f"matcher {matcher} did not match; resource '{resource_id}' not forwarded")

    return forward_series



import logging
import threading
from datetime import datetime

ONE_MINUTE = 60
ONE_HOUR = ONE_MINUTE * 60
ONE_DAY = 24 * ONE_HOUR


def handle_failed_forwarding_job(first_retry=ONE_MINUTE, job_types=('DicomModalityStore',)):
    def handle_failed_forwarding_job(event, session):
        job_id = event.resource_id
        resp = session.get(f'/jobs/{job_id}')
        resp.raise_for_status()

        job = resp.json()
        job_type = job['Type']
        if job_type not in job_types:
            logging.debug(f"not retrying '{job_type}' job '{job_id}'")
            return

        delay = calculate_delay(job, first_retry)
        logging.debug(f"resubmitting job '{job_id}' after {delay} seconds")
        timer = threading.Timer(interval=delay, function=resubmit_job, args=[session, job_id, delay])
        timer.start()

    return handle_failed_forwarding_job


def calculate_delay(job, first_retry=ONE_MINUTE):
    elapsed = parse_time(job['CompletionTime']) - parse_time(job['CreationTime'])
    return min(max(first_retry, elapsed.seconds * 2), ONE_DAY)


def parse_time(job_time):
    return datetime.strptime(job_time, '%Y%m%dT%H%M%S.%f')


def resubmit_job(session, job_id, delay):
    resp = session.post(f'/jobs/{job_id}/resubmit')
    resp.raise_for_status()
    logging.info(f'resubmitted job {job_id}')



import json
import logging

from dataclasses import dataclass

from orthanc_ext.logging_configurator import orthanc_logging
from orthanc_ext.python_utilities import ensure_iterable, create_reverse_type_dict
from orthanc_ext.requests_utilities import create_internal_requests_session, get_rest_api_base_url, get_certificate


def register_event_handlers(event_handlers, orthanc_module, requests_session, logging_configuration=orthanc_logging):
    logging_configuration(orthanc_module)

    @dataclass
    class ChangeEvent:
        change_type: int
        resource_type: int
        resource_id: str

        def __str__(self):
            return (
                f'ChangeEvent(change_type={event_types.get(self.change_type)}, '
                f"resource_type={resource_types.get(self.resource_type)}, resource_id='{self.resource_id}')"
            )

    def create_type_index(orthanc_type):
        return create_reverse_type_dict(orthanc_type)

    event_types = create_type_index(orthanc_module.ChangeType)
    resource_types = create_type_index(orthanc_module.ResourceType)

    event_handlers = {k: ensure_iterable(v) for k, v in event_handlers.items()}

    def unhandled_event_logger(event, _):
        logging.debug(f'no handler registered for {event_types[event.change_type]}')

    def OnChange(change_type, resource_type, resource_id):
        handlers = event_handlers.get(change_type, [unhandled_event_logger])
        return_values = []
        for handler in handlers:
            event = ChangeEvent(change_type, resource_type, resource_id)
            return_values.append(handler(event, requests_session))
        return return_values

    orthanc_module.RegisterOnChangeCallback(OnChange)


def create_session(orthanc):
    config = json.loads(orthanc.GetConfiguration())
    return create_internal_requests_session(
        get_rest_api_base_url(config), orthanc.GenerateRestApiAuthorizationToken(), get_certificate(config)
    )



"""
This module implements the Orthanc Python plugin API to run requests against an external Orthanc instance.

This will allow you quickly evolve your python scripts and make them easy to integration test as well.
"""
import uuid


class OrthancApiHandler(object):
    class ResourceType:
        PATIENT = 0
        STUDY = 1
        SERIES = 2
        INSTANCE = 3
        NONE = 4

    # redefine to make this type available for unit tests outside the Orthanc Python plugin.
    # https://hg.orthanc-server.com/orthanc-python/file/tip/Sources/Autogenerated/sdk_OrthancPluginChangeType.impl.h
    class ChangeType:
        COMPLETED_SERIES = 0
        DELETED = 1
        NEW_CHILD_INSTANCE = 2
        NEW_INSTANCE = 3
        NEW_PATIENT = 4
        NEW_SERIES = 5
        NEW_STUDY = 6
        STABLE_PATIENT = 7
        STABLE_SERIES = 8
        STABLE_STUDY = 9
        ORTHANC_STARTED = 10
        ORTHANC_STOPPED = 11
        UPDATED_ATTACHMENT = 12
        UPDATED_METADATA = 13
        UPDATED_PEERS = 14
        UPDATED_MODALITIES = 15
        JOB_SUBMITTED = 16
        JOB_SUCCESS = 17
        JOB_FAILURE = 18

        # not defined by orthanc
        UNKNOWN = 999

    @staticmethod
    def GenerateRestApiAuthorizationToken():
        return uuid.uuid4()

    def RegisterOnChangeCallback(self, change_callback):
        self.change_callback = change_callback

    def on_change(self, change_type, resource_type, resource_id):
        return self.change_callback(change_type, resource_type, resource_id)

    def LogInfo(self, message):
        print(f'INFO: {message}')

    def LogWarning(self, message):
        print(f'WARNING: {message}')

    def LogError(self, message):
        print(f'ERROR: {message}')



import logging
from enum import Enum


def python_logging(_, default_level=logging.INFO):
    """Configures python logging.
    Useful when Orthanc is using stderr and stdout handlers: offers more log levels and a better date format"""
    fmt = '%(levelname)s %(asctime)s %(filename)s:%(lineno)s] %(message)s'
    logging.basicConfig(format=fmt)
    logger = logging.getLogger()
    logger.setLevel(default_level)
    return fmt


def orthanc_logging(orthanc_module, default_level=logging.INFO):
    """Configures orthanc logging. Useful when orthanc is configured to write to a log file"""
    logger = logging.getLogger()
    logger.setLevel(default_level)
    logger.addHandler(OrthancLogHandler(orthanc_module))


class OrthancLogHandler(logging.Handler):
    def __init__(self, orthanc_module):
        logging.Handler.__init__(self)
        self.orthanc_module = orthanc_module
        self.log_func_mapping = {
            logging.INFO: orthanc_module.LogInfo,
            logging.WARNING: orthanc_module.LogWarning,
            logging.ERROR: orthanc_module.LogError,
            logging.CRITICAL: orthanc_module.LogError,
        }

    def emit(self, record: logging.LogRecord) -> None:
        self.log_func_mapping.get(record.levelno, self.orthanc_module.LogInfo)(
            logging.Formatter(fmt='[%(filename)s:%(lineno)s] %(message)s').format(record)
        )


class OrthancLevel(Enum):
    DEFAULT = ('default', 'WARNING')
    VERBOSE = ('verbose', 'INFO')
    TRACE = ('trace', 'DEBUG')

    def __init__(self, orthanc_level, python_level):
        self.orthanc_level = orthanc_level
        self.python_level = python_level


def configure_log_level(session, level: OrthancLevel):
    session.put('/tools/log-level-plugins', level.value)



from orthanc_ext.orthanc import OrthancApiHandler

orthanc = OrthancApiHandler()


def test_GenerateRestApiAuthorizationToken_should_yield_a_token():
    assert orthanc.GenerateRestApiAuthorizationToken() is not None



import logging

import responses

from orthanc_ext import event_dispatcher
from orthanc_ext.logging_configurator import python_logging
from orthanc_ext.orthanc import OrthancApiHandler
from orthanc_ext.orthanc_utilities import get_metadata_of_first_instance_of_series
from orthanc_ext.scripts.auto_forward import forward_dicom, DicomReceivedMatcher

orthanc = OrthancApiHandler()
session = event_dispatcher.create_internal_requests_session('https://localhost:8042', '')


@responses.activate
def test_autoforward_on_match_shall_start_start_modality_store(caplog):
    responses.add(responses.POST, 'https://localhost:8042/modalities/pacs/store', body='study-uuid', status=200)

    register_and_trigger_handler([DicomReceivedMatcher(lambda uid, _: True, lambda uid, _: 'pacs')])

    assert caplog.messages == ["DICOM export to modality 'pacs' started for resource 'study-uuid'"]


def is_not_dicom_origin(resource_id, session):
    return get_metadata_of_first_instance_of_series(session, resource_id, 'Origin') != 'DicomProtocol'


@responses.activate
def test_autoforward_on_multiple_matches_shall_start_start_modality_store(caplog):
    caplog.set_level(logging.INFO)
    responses.add(
        responses.GET,
        'https://localhost:8042/series/study-uuid/instances',
        json=[{'ID': 'b99cd218-ae67f0d7-70324b6b-2b095801-f858dedf'}],
    )
    responses.add(
        responses.GET,
        'https://localhost:8042/instances/b99cd218-ae67f0d7-70324b6b-2b095801-f858dedf/metadata/Origin',
        body='Plugins',
    )
    responses.add(responses.POST, 'https://localhost:8042/modalities/pacs1/store', body='study-uuid', status=200)
    responses.add(responses.POST, 'https://localhost:8042/modalities/pacs2/store', body='study-uuid', status=200)

    matcher1 = DicomReceivedMatcher(is_not_dicom_origin, lambda uid, _: 'pacs1')
    matcher2 = DicomReceivedMatcher(lambda uid, _: True, lambda uid, _: 'pacs2')

    register_and_trigger_handler([matcher1, matcher2])

    assert caplog.messages == [
        "DICOM export to modality 'pacs1' started for resource 'study-uuid'",
        "DICOM export to modality 'pacs2' started for resource 'study-uuid'",
    ]


def register_and_trigger_handler(matchers):
    event_dispatcher.register_event_handlers(
        {orthanc.ChangeType.STABLE_STUDY: forward_dicom(matchers)},
        orthanc_module=orthanc,
        requests_session=session,
        logging_configuration=python_logging,
    )
    orthanc.on_change(orthanc.ChangeType.STABLE_STUDY, '', 'study-uuid')


def test_autoforward_on_no_match_shall_log_and_continue(caplog):
    register_and_trigger_handler([DicomReceivedMatcher(lambda uid, _: False, lambda uid, _: 'pacs')])

    (message,) = caplog.messages
    assert "did not match; resource 'study-uuid' not forwarded" in message



"""
Convenience methods to work with the Orthanc REST API
"""


def get_metadata_of_first_instance_of_series(session, series_id, metadata_key):
    instances = session.get(f'/series/{series_id}/instances').json()
    assert len(instances) > 0, f'expected at least one instance in series {series_id}'
    resp = session.get(f"/instances/{instances[0]['ID']}/metadata/{metadata_key}")
    resp.raise_for_status()

    return resp.text



import logging
import re

from orthanc_ext.logging_configurator import python_logging
from orthanc_ext.orthanc import OrthancApiHandler


def test_setup_orthanc_log_format(caplog):
    fmt = python_logging(OrthancApiHandler())
    # pytest overrides the log configuration; need to explicitly set the formatter to test it
    caplog.handler.setFormatter(logging.Formatter(fmt))
    logging.info('message')

    # level names and time format deviate from orthanc; timestamp includes date
    assert re.match(r'INFO 20\d\d-\d\d-\d\d .* test_logging_configurator.py:\d\d] message', caplog.text)



import logging

import requests
import responses
from attr import dataclass

from orthanc_ext import event_dispatcher
from orthanc_ext.logging_configurator import python_logging
from orthanc_ext.orthanc import OrthancApiHandler

orthanc = OrthancApiHandler()


@dataclass
class ChangeEvent:
    change: orthanc.ChangeType = orthanc.ChangeType.UNKNOWN
    resource_type: int = orthanc.ResourceType.NONE
    resource_id: str = None
    orthanc = None


def capture(event):
    def capture_impl(incoming_event, local_orthanc):
        event.change = incoming_event.change_type
        event.resource_type = incoming_event.resource_type
        event.resource_id = incoming_event.resource_id
        event.orthanc = local_orthanc

    return capture_impl


def test_registered_callback_should_be_triggered_on_change_event():
    event = ChangeEvent()

    event_dispatcher.register_event_handlers(
        {orthanc.ChangeType.STABLE_STUDY: capture(event)}, orthanc_module=orthanc, requests_session=requests
    )

    orthanc.on_change(orthanc.ChangeType.STABLE_STUDY, orthanc.ResourceType.STUDY, 'resource-uuid')

    assert event.resource_id == 'resource-uuid'
    assert event.resource_type == orthanc.ResourceType.STUDY
    assert event.orthanc is not None


def test_all_registered_callbacks_should_be_triggered_on_change_event():
    event1 = ChangeEvent()
    event2 = ChangeEvent()

    event_dispatcher.register_event_handlers(
        {orthanc.ChangeType.STABLE_STUDY: [capture(event1), capture(event2)]},
        orthanc_module=orthanc,
        requests_session=requests,
    )

    orthanc.on_change(orthanc.ChangeType.STABLE_STUDY, orthanc.ResourceType.STUDY, 'resource-uuid')

    assert event1.resource_id is not None
    assert event2.resource_id is not None


def test_no_registered_callbacks_should_be_reported_in_on_change_event(caplog):

    event_dispatcher.register_event_handlers(
        {}, orthanc_module=orthanc, requests_session=requests, logging_configuration=python_logging
    )
    caplog.set_level(logging.DEBUG)
    orthanc.on_change(orthanc.ChangeType.ORTHANC_STARTED, '', '')

    assert 'no handler registered for ORTHANC_STARTED' in caplog.text


@responses.activate
def test_shall_return_values_from_executed_handlers():
    responses.add(responses.GET, 'http://localhost:8042/system', json={'Version': '1.9.0'})

    def get_system_info(_, session):
        return session.get('http://localhost:8042/system').json()

    event_dispatcher.register_event_handlers(
        {orthanc.ChangeType.ORTHANC_STARTED: get_system_info}, orthanc_module=orthanc, requests_session=requests
    )
    (system_info,) = orthanc.on_change(orthanc.ChangeType.ORTHANC_STARTED, orthanc.ResourceType.NONE, '')
    assert system_info.get('Version') == '1.9.0'


def test_event_shall_have_human_readable_representation(caplog):
    caplog.set_level(logging.INFO)

    def log_event(evt, _):
        logging.info(evt)

    event_dispatcher.register_event_handlers(
        {orthanc.ChangeType.STABLE_STUDY: log_event}, orthanc_module=orthanc, requests_session=requests
    )
    orthanc.on_change(orthanc.ChangeType.STABLE_STUDY, orthanc.ResourceType.STUDY, 'uuid')

    assert 'change_type=STABLE_STUDY' in caplog.text
    assert 'resource_type=STUDY' in caplog.text



from requests_toolbelt import sessions


def get_rest_api_base_url(config):
    port = config.get('HttpPort', 8042)
    scheme = 'https' if config.get('SslEnabled', False) else 'http'
    return f'{scheme}://localhost:{port}/'


def get_certificate(config):
    return config.get('SslCertificate', False)


def create_internal_requests_session(base_url, token, cert=False):
    session = sessions.BaseUrlSession(base_url)
    session.headers['Authorization'] = token
    session.verify = cert
    return session



from typing import Iterable


def ensure_iterable(v):
    return v if isinstance(v, Iterable) else [v]


def hashable(k):
    try:
        return hash(k)
    except TypeError:
        return False


def create_reverse_type_dict(py_type):
    return {v: k for k, v in py_type.__dict__.items() if hashable(v)}



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['requests', 'requests_toolbelt', 'dataclasses']

setup_requirements = [
    'pytest-runner',
]

test_requirements = ['pytest>=3', 'responses']

setup(
    author='WalkIT',
    author_email='code@walkit.nl',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="A simple Orthanc python plugin based framework to extend Orthanc's feature set with testable python "
    'scripts.',
    install_requires=requirements,
    license='GNU Affero General Public License v3',
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='orthanc testing',
    name='orthanc-server-extensions',
    packages=find_packages(include=['orthanc_ext', 'orthanc_ext.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/walkIT-nl/orthanc-server-extensions',
    version='3.1.9',
    zip_safe=False,
)



import logging

import responses

from orthanc_ext import event_dispatcher
from orthanc_ext.logging_configurator import python_logging
from orthanc_ext.orthanc import OrthancApiHandler
from orthanc_ext.scripts.auto_retries import handle_failed_forwarding_job, calculate_delay, ONE_MINUTE, ONE_DAY

orthanc = OrthancApiHandler()
session = event_dispatcher.create_internal_requests_session('https://localhost:8042', '')


def test_calculate_delay():
    job = {
        'CompletionTime': '20210210T084933.795611',
        'Content': {
            'Description': 'REST API',
            'FailedInstancesCount': 0,
            'InstancesCount': 1,
            'LocalAet': 'ORTHANC',
            'ParentResources': ['3121d449-9b15610c-df9b8396-bee611db-3901f794'],
            'RemoteAet': 'PYNETDICOM',
        },
        'CreationTime': '20210210T084350.430751',
        'EffectiveRuntime': 0.036999999999999998,
        'ErrorCode': 9,
        'ErrorDescription': 'Error in the network protocol',
        'ID': '0a9b0d5f-a2a8-46c1-8b2a-6a1e081427fb',
        'Priority': 0,
        'Progress': 0,
        'State': 'Failure',
        'Timestamp': '20210210T090925.594915',
        'Type': 'DicomModalityStore',
    }

    assert calculate_delay(job) == 686


def test_calculate_delay_shall_not_retry_too_aggressively():
    # interval first try: 1 second
    job = {'CreationTime': '20210210T084350.430751', 'CompletionTime': '20210210T084351.430751'}

    assert calculate_delay(job) == ONE_MINUTE


def test_calculate_delay_shall_use_back_off():
    # time between previous tries: 3 minutes
    job = {'CreationTime': '20210210T084350.430751', 'CompletionTime': '20210210T084650.430751'}

    assert calculate_delay(job) == 6 * ONE_MINUTE

    job = {'CreationTime': '20210210T084350.430751', 'CompletionTime': '20210210T085250.430751'}

    assert calculate_delay(job) == 18 * ONE_MINUTE


def test_calculate_delay_shall_retry_every_day():
    job = {'CreationTime': '20210210T084350.430751', 'CompletionTime': '20210210T224350.430751'}

    assert calculate_delay(job) == ONE_DAY


@responses.activate
def test_should_not_resubmit_other_job_types(caplog):
    job = {
        'CreationTime': '20210210T084350.430751',
        'CompletionTime': '20210210T224350.430751',
        'Type': 'CreateDicomZip',
    }

    responses.add(responses.GET, 'https://localhost:8042/jobs/job-uuid', json=job)

    event_dispatcher.register_event_handlers(
        {
            orthanc.ChangeType.JOB_FAILURE: handle_failed_forwarding_job(0.1),
        },
        orthanc_module=orthanc,
        requests_session=session,
        logging_configuration=python_logging,
    )
    caplog.set_level(logging.DEBUG)

    orthanc.on_change(orthanc.ChangeType.JOB_FAILURE, '', 'job-uuid')

    assert caplog.messages == ["not retrying 'CreateDicomZip' job 'job-uuid'"]


@responses.activate
def test_on_failure_should_resubmit_job(caplog):
    responses.add(
        responses.GET,
        'https://localhost:8042/jobs/job-uuid',
        json={
            'CreationTime': '20210210T084350.430751',
            'CompletionTime': '20210210T084351.430751',
            'Type': 'DicomModalityStore',
        },
    )

    # TODO: figure out why expectation not met in Timer thread.
    responses.add(responses.POST, 'https://localhost:8042/jobs/job-uuid/resubmit', {})

    event_dispatcher.register_event_handlers(
        {
            orthanc.ChangeType.JOB_FAILURE: handle_failed_forwarding_job(0.1),
        },
        orthanc_module=orthanc,
        requests_session=session,
        logging_configuration=python_logging,
    )

    caplog.set_level(logging.DEBUG)
    orthanc.on_change(orthanc.ChangeType.JOB_FAILURE, '', 'job-uuid')

    assert caplog.messages == ["resubmitting job 'job-uuid' after 2 seconds"]



"""Test entry point script for Orthanc Python Plugin."""

import logging

# provided by the plugin runtime.
import orthanc

import orthanc_ext.event_dispatcher
from orthanc_ext import event_dispatcher


def log_event(param):
    def log_event_impl(event, _):
        logging.info(f"orthanc '{event}' event handled with param {param}")

    return log_event_impl


def start_maintenance_cycle(event, _):
    logging.info(f'do something special on {event}')


def show_system_info(_, session):
    logging.warning('orthanc version retrieved: %s', session.get('/system').json().get('Version'))


event_dispatcher.register_event_handlers(
    {
        orthanc.ChangeType.ORTHANC_STARTED: [log_event('started'), start_maintenance_cycle, show_system_info],
        orthanc.ChangeType.ORTHANC_STOPPED: log_event('stopped'),
    },
    orthanc,
    orthanc_ext.event_dispatcher.create_session(orthanc),
)
